// This file is taken from [`chili`]

use std::{
    cell::UnsafeCell,
    ptr::NonNull,
    sync::{
        Arc,
        atomic::{AtomicU8, AtomicU32, Ordering},
    },
    thread,
};

enum State {
    Pending,
    Waiting,
    Ready,
    Taken,
}

// taken from `std`
#[derive(Debug)]
#[repr(transparent)]
pub struct Parker {
    mutex: AtomicU32,
}

impl Parker {
    const PARKED: u32 = u32::MAX;
    const EMPTY: u32 = 0;
    const NOTIFIED: u32 = 1;
    pub fn new() -> Self {
        Self {
            mutex: AtomicU32::new(Self::EMPTY),
        }
    }

    pub fn is_parked(&self) -> bool {
        self.mutex.load(Ordering::Acquire) == Self::PARKED
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all, fields(this = self as *const Self as usize)))]
    pub fn park(&self) {
        if self.mutex.fetch_sub(1, Ordering::Acquire) == Self::NOTIFIED {
            // The thread was notified, so we can return immediately.
            return;
        }

        loop {
            atomic_wait::wait(&self.mutex, Self::PARKED);

            // We check whether we were notified or woke up spuriously with
            // acquire ordering in order to make-visible any writes made by the
            // thread that notified us.
            if self.mutex.swap(Self::EMPTY, Ordering::Acquire) == Self::NOTIFIED {
                // The thread was notified, so we can return immediately.
                return;
            } else {
                // spurious wakeup, so we need to re-park.
                continue;
            }
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all, fields(this = self as *const Self as usize)))]
    pub fn unpark(&self) {
        // write with Release ordering to ensure that any writes made by this
        // thread are made-available to the unparked thread.
        if self.mutex.swap(Self::NOTIFIED, Ordering::Release) == Self::PARKED {
            // The thread was parked, so we need to notify it.
            atomic_wait::wake_one(&self.mutex);
        } else {
            // The thread was not parked, so we don't need to do anything.
        }
    }
}

#[derive(Debug)]
#[repr(C)]
struct Channel<T = ()> {
    state: AtomicU8,
    /// Can only be written only by the `Receiver` and read by the `Sender` if
    /// `state` is `State::Waiting`.
    waiting_thread: NonNull<Parker>,
    /// Can only be written only by the `Sender` and read by the `Receiver` if
    /// `state` is `State::Ready`.
    val: UnsafeCell<Option<Box<thread::Result<T>>>>,
}

impl<T> Channel<T> {
    fn new(waiting_thread: NonNull<Parker>) -> Self {
        Self {
            state: AtomicU8::new(State::Pending as u8),
            waiting_thread,
            val: UnsafeCell::new(None),
        }
    }
}

#[derive(Debug)]
pub struct Receiver<T = ()>(Arc<Channel<T>>);

impl<T: Send> Receiver<T> {
    pub fn is_empty(&self) -> bool {
        self.0.state.load(Ordering::Acquire) != State::Ready as u8
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn wait(&self) {
        loop {
            match self.0.state.compare_exchange(
                State::Pending as u8,
                State::Waiting as u8,
                Ordering::AcqRel,
                Ordering::Acquire,
            ) {
                Ok(_) => {
                    // SAFETY:
                    // The `waiting_thread` is set to the current thread's parker
                    // before we park it.
                    unsafe {
                        let thread = self.0.waiting_thread.as_ref();
                        thread.park();
                    }

                    // we might have been woken up because of a shared job.
                    // In that case, we need to check the state again.
                    if self
                        .0
                        .state
                        .compare_exchange(
                            State::Waiting as u8,
                            State::Pending as u8,
                            Ordering::AcqRel,
                            Ordering::Acquire,
                        )
                        .is_ok()
                    {
                        continue;
                    }
                }
                Err(state) if state == State::Ready as u8 => {
                    // The channel is ready, so we can return immediately.
                    return;
                }
                _ => {
                    panic!("Receiver is already waiting or consumed.");
                }
            }
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn poll(&self) -> Option<thread::Result<T>> {
        if self
            .0
            .state
            .compare_exchange(
                State::Ready as u8,
                State::Taken as u8,
                Ordering::AcqRel,
                Ordering::Acquire,
            )
            .is_ok()
        {
            unsafe { Some(self.take()) }
        } else {
            None
        }
    }

    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn recv(self) -> thread::Result<T> {
        self.wait();

        // SAFETY:
        // To arrive here, either `state` is `State::Ready` or the above
        // `compare_exchange` succeeded, the thread was parked and then
        // unparked by the `Sender` *after* the `state` was set to
        // `State::Ready`.
        //
        // In either case, this thread now has unique access to `val`.
        unsafe { self.take() }
    }

    unsafe fn take(&self) -> thread::Result<T> {
        let result = unsafe { (*self.0.val.get()).take().map(|b| *b).unwrap() };

        assert_eq!(
            self.0.state.swap(State::Taken as u8, Ordering::Release),
            State::Ready as u8
        );

        result
    }
}

#[derive(Debug)]
#[repr(transparent)]
pub struct Sender<T = ()>(Arc<Channel<T>>);

impl<T: Send> Sender<T> {
    #[cfg_attr(feature = "tracing", tracing::instrument(level = "trace", skip_all))]
    pub fn send(self, val: thread::Result<T>) {
        // SAFETY:
        // Only this thread can write to `val` and none can read it
        // yet.
        unsafe {
            *self.0.val.get() = Some(Box::new(val));
        }

        if self.0.state.swap(State::Ready as u8, Ordering::AcqRel) == State::Waiting as u8 {
            // SAFETY:
            // A `Receiver` already wrote its thread to `waiting_thread`
            // *before* setting the `state` to `State::Waiting`.
            unsafe {
                let thread = self.0.waiting_thread.as_ref();
                thread.unpark();
            }
        }
    }
}

pub fn channel<T: Send>(thread: NonNull<Parker>) -> (Sender<T>, Receiver<T>) {
    let channel = Arc::new(Channel::new(thread));

    (Sender(channel.clone()), Receiver(channel))
}
