///! Rayon's job logic
use std::{cell::UnsafeCell, marker::PhantomPinned};

pub mod spice;
pub mod v2;

pub trait Job<Args = ()> {
    unsafe fn execute(this: *const (), args: Args);
}

pub struct JobRef<Args = ()> {
    this: *const (),
    execute_fn: unsafe fn(*const (), Args),
}

unsafe impl<Args> Send for JobRef<Args> {}
unsafe impl<Args> Sync for JobRef<Args> {}

impl<Args> JobRef<Args> {
    pub unsafe fn new<T>(data: *const T) -> JobRef<Args>
    where
        T: Job<Args>,
    {
        Self {
            this: data.cast(),
            execute_fn: <T as Job<Args>>::execute,
        }
    }

    pub fn id(&self) -> impl Eq {
        (self.this, self.execute_fn)
    }

    pub unsafe fn execute(self, args: Args) {
        unsafe { (self.execute_fn)(self.this, args) }
    }
}

pub struct StackJob<F> {
    task: UnsafeCell<Option<F>>,
    _phantom: PhantomPinned,
}

impl<F> StackJob<F> {
    pub fn new(task: F) -> StackJob<F> {
        Self {
            task: UnsafeCell::new(Some(task)),
            _phantom: PhantomPinned,
        }
    }

    pub unsafe fn take_once(self) -> F {
        self.task.into_inner().unwrap()
    }

    #[inline]
    pub fn run<Args>(self, args: Args)
    where
        F: FnOnce(Args),
    {
        self.task.into_inner().unwrap()(args);
    }

    #[inline]
    pub unsafe fn as_task_ref<Args>(&self) -> JobRef<Args>
    where
        F: FnOnce(Args),
    {
        unsafe { JobRef::<Args>::new(self) }
    }
}

impl<Args, F> Job<Args> for StackJob<F>
where
    F: FnOnce(Args),
{
    unsafe fn execute(this: *const (), args: Args) {
        let this = &*this.cast::<Self>();
        let func = (*this.task.get()).take().unwrap();
        func(args);
        // set internal latch here?
    }
}

pub struct HeapJob<F = ()>
where
    F: Send,
{
    func: F,
    _phantom: PhantomPinned,
}

impl<F> HeapJob<F>
where
    F: Send,
{
    pub fn new(task: F) -> Box<HeapJob<F>> {
        Box::new(Self {
            func: task,
            _phantom: PhantomPinned,
        })
    }

    #[inline]
    pub unsafe fn into_static_task_ref<Args>(self: Box<Self>) -> JobRef<Args>
    where
        F: FnOnce(Args) + 'static,
    {
        self.into_task_ref()
    }

    #[inline]
    pub unsafe fn into_task_ref<Args>(self: Box<Self>) -> JobRef<Args>
    where
        F: FnOnce(Args),
    {
        JobRef::new(Box::into_raw(self))
    }
}

impl<Args, F> Job<Args> for HeapJob<F>
where
    F: FnOnce(Args) + Send,
{
    unsafe fn execute(this: *const (), args: Args) {
        let this = Box::from_raw(this.cast::<Self>().cast_mut());
        (this.func)(args);
        // set internal latch here?
    }
}
