use std::{
    mem::{self, MaybeUninit},
    pin::{pin, Pin},
};

use tracing_test::traced_test;

use super::{util::TaggedAtomicPtr, *};

fn pin_ptr<T>(pin: &Pin<&mut T>) -> NonNull<T> {
    let a = &raw const **pin;
    unsafe { NonNull::new_unchecked(a.cast_mut()) }
}

#[test]
fn new_job() {
    let mut list = JobList::new();

    let stack = StackJob::new(|| 3 + 3, crate::latch::MutexLatch::new());

    let job = stack.as_job();
    unsafe {
        list.push_front(&job);
    }

    unsafe {
        let job_ref = list.pop_front().unwrap().cast::<Job<i32>>().as_ref();
        job_ref.set_pending();

        _ = stack.unwrap();
        job_ref.complete(Ok(6));
        let result = job_ref.wait().into_inner();
        assert_eq!(result.ok(), Some(6));
    }
}

#[test]
fn job_list_pop_back() {
    let mut list = JobList::new();
    let a = pin!(Job::empty());
    let b = pin!(Job::empty());
    let c = pin!(Job::empty());

    unsafe {
        list.push_front(&*a);
        list.push_front(&*b);
        list.push_back(&*c);
    }

    assert_eq!(list.pop_back(), Some(pin_ptr(&c)));
    unsafe {
        list.push_front(&*c);
    }
    assert_eq!(list.pop_back(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_back(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_back(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_front(), None);
}

#[test]
fn job_list_pop_front() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());
    let b = pin!(Job::<()>::empty());
    let c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
        list.push_front(&*b);
        list.push_back(&*c);
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    unsafe {
        list.push_back(&*b);
    }
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_middle() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());
    let b = pin!(Job::<()>::empty());
    let c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
        list.push_front(&*b);
        list.push_front(&*c);
    }

    unsafe {
        b.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_front() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());
    let b = pin!(Job::<()>::empty());
    let c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
        list.push_front(&*b);
        list.push_front(&*c);
    }

    unsafe {
        a.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&c)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_back() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());
    let b = pin!(Job::<()>::empty());
    let c = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
        list.push_front(&*b);
        list.push_front(&*c);
    }

    unsafe {
        c.unlink();
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&b)));
    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn unlink_job_single() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
    }

    unsafe {
        a.unlink();
    }

    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn tagged_ptr_exchange() {
    let boxed = Box::into_raw(Box::new(42usize));
    let _guard = DropGuard::new(|| drop(unsafe { Box::from_raw(boxed) }));

    let ptr = TaggedAtomicPtr::<_, 3>::new(boxed, 1usize);

    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
    ptr.set_tag(3, Ordering::Release, Ordering::Relaxed);
    assert_eq!(ptr.tag(Ordering::Relaxed), 3);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);

    let old = ptr.compare_exchange_tag(3, 4, Ordering::Release, Ordering::Relaxed);

    assert_eq!(old, Ok(3));
    assert_eq!(ptr.tag(Ordering::Relaxed), 4);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
}

#[test]
fn tagged_ptr_exchange_failure() {
    let boxed = Box::into_raw(Box::new(42usize));
    let _guard = DropGuard::new(|| drop(unsafe { Box::from_raw(boxed) }));

    let ptr = TaggedAtomicPtr::<_, 3>::new(boxed, 1usize);

    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);

    let old = ptr.compare_exchange_tag(3, 4, Ordering::Release, Ordering::Relaxed);
    assert_eq!(old, Err(1));
    assert_eq!(ptr.tag(Ordering::Relaxed), 1);
    assert_eq!(ptr.ptr(Ordering::Relaxed).as_ptr(), boxed);
}

#[test]
fn value_inline() {
    let val = SmallBox::new(3usize);

    let inner = val.into_inner();
    assert_eq!(inner, 3usize);
}

#[test]
fn value_inline_struct() {
    #[derive(Default, PartialEq, Debug)]
    struct Small {
        c: f32,
    }
    let val = SmallBox::new(Small { c: 3.2 });

    let inner = val.into_inner();
    assert_eq!(inner.c, 3.2);
}

#[test]
fn value_inline_ptr_sized() {
    assert!(SmallBox::<usize>::is_inline());
}

#[test]
fn value_boxed() {
    #[derive(Default, PartialEq, Debug)]
    struct Big {
        a: usize,
        b: f32,
        c: u32,
    }
    let val = SmallBox::new(Big {
        a: 42,
        b: 2.25,
        c: 7,
    });

    let inner = val.into_inner();
    assert_eq!(
        inner,
        Big {
            a: 42,
            b: 2.25,
            c: 7
        }
    );
}

#[test]
fn value_inline_drop() {
    #[derive(PartialEq, Debug)]
    #[repr(transparent)]
    struct Small<'a> {
        inner: &'a mut usize,
    }
    impl Drop for Small<'_> {
        fn drop(&mut self) {
            *self.inner += 1;
        }
    }
    assert!(SmallBox::<Small<'_>>::is_inline());
    let mut dropped = 0;
    {
        let inner = {
            let val = SmallBox::new(Small {
                inner: &mut dropped,
            });

            val.into_inner()
        };
        assert_eq!(*inner.inner, 0);
    }
    assert_eq!(dropped, 1);
    {
        let _val = SmallBox::new(Small {
            inner: &mut dropped,
        });
    }
    assert_eq!(dropped, 2);
}

#[test]
fn value_boxed_drop() {
    #[derive(PartialEq, Debug)]
    struct Big<'a> {
        inner: &'a mut usize,
        pad: [usize; 3],
    }
    impl Drop for Big<'_> {
        fn drop(&mut self) {
            *self.inner += 1;
        }
    }
    let mut dropped = 0;
    {
        let inner = {
            let val = SmallBox::new(Big {
                inner: &mut dropped,
                pad: [0; 3],
            });

            val.into_inner()
        };
        assert_eq!(*inner.inner, 0);
    }
    assert_eq!(dropped, 1);
    {
        let _val = SmallBox::new(Big {
            inner: &mut dropped,
            pad: [0; 3],
        });
    }
    assert_eq!(dropped, 2);
}

#[test]
fn job_list_pop_front_empty() {
    let mut list = JobList::new();

    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_front(), None);
}

#[test]
fn job_list_pop_back_empty() {
    let mut list = JobList::new();

    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn job_list_pop_back_emptied() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
    }

    assert_eq!(list.pop_back(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_back(), None);
    assert_eq!(list.pop_back(), None);
}

#[test]
fn job_list_pop_front_emptied() {
    let mut list = JobList::new();
    let a = pin!(Job::<()>::empty());

    unsafe {
        list.push_front(&*a);
    }

    assert_eq!(list.pop_front(), Some(pin_ptr(&a)));
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_front(), None);
    assert_eq!(list.pop_front(), None);
}

#[test]
fn spawn() {
    let pool = ThreadPool::new();

    let mut x = 0;
    pool.scope(|s| {
        tracing::info!("scope");
        s.spawn(|_| {
            tracing::info!("spawn");
            x += 1;
            tracing::info!("x: {x}");
        });
        tracing::info!("~scope");
    });

    eprintln!("x: {x}");
}

#[test]
fn spawn_borrow() {
    let pool = ThreadPool::new();

    pool.scope(|s| {
        let mut x = 0;
        s.spawn(|_| {
            x += 1;
            assert_eq!(x, 1);
        });
    });
}

#[test]
fn spawn_future() {
    let pool = ThreadPool::new();

    let task = pool.scope(|s| {
        let task = s.spawn_future(async {
            eprintln!("executing future");
            3 + 1
        });

        // let task = s.spawn_future(async {
        //     eprintln!("task: {}", task.await);
        // });
        task
    });
    let x = async_std::task::block_on(task);
    eprintln!("x: {x}");
}

#[test]
fn join() {
    let pool = ThreadPool::new();

    let x = pool.scope(|s| {
        let (a, b) = s.join(|_| 3, |_| 4);

        a + b
    });

    eprintln!("x: {x}");
}

#[test]
fn join_many() {
    use crate::util::tree::Tree;

    let pool = ThreadPool::new();

    let tree = Tree::new(4, 1u32);

    fn sum(tree: &Tree<u32>, node: usize, scope: &Scope) -> u32 {
        let node = tree.get(node);
        let (l, r) = scope.join(
            |s| node.left.map(|node| sum(tree, node, s)).unwrap_or_default(),
            |s| {
                node.right
                    .map(|node| sum(tree, node, s))
                    .unwrap_or_default()
            },
        );

        // eprintln!("node: {node:?}, l: {l}, r: {r}");

        node.leaf + l + r
    }

    let sum = pool.scope(|s| sum(&tree, tree.root().unwrap(), s));
    eprintln!("{sum}");
}

#[test]
fn rebox() {
    struct A(u32);

    impl Drop for A {
        fn drop(&mut self) {
            eprintln!("drop");
        }
    }

    let ptr = Box::into_raw(Box::new(A(5)));

    let x = unsafe { &*ptr.offset(mem::offset_of!(A, 0) as isize).cast::<u32>() };
    let y = *x;

    unsafe {
        ptr.drop_in_place();
        _ = Box::<MaybeUninit<A>>::from_raw(ptr.cast());
    }
    assert_eq!(y, 5);
}
